/*\
|*| --------------------------------------------------------------------------
|*|     KGI kernel driver
|*| --------------------------------------------------------------------------
|*|
|*|	Copyright (C) 1995 by 
|*|
|*|	  Andreas Beck,		[becka@sunserver1.rz.uni-duesseldorf.de]
|*|	  Steffen Seeger	[seeger@physik.tu-chemnitz.de]
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	This program is distributed in the hope that it will be useful,
|*|	but WITHOUT ANY WARRANTY; without even the implied warranty of
|*|	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
|*|	GNU General Public License for more details.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ---------------------------------------------------------------------------
\*/

#undef DEBUG

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/malloc.h>

#include "graphdev.h"
#include "kgi.h"

#define	MAX_NR_CONSOLES	8
#define	TTY_NR	MINOR(inode->i_rdev)
#define	VC_GRAPHMODE	32

char kernel_version[] = UTS_RELEASE;

static struct kgi_mode textmode;		   /* text mode for VC's   */

#define	NO_MODE_ASSIGNED(X)	(tty[X].gr_mode.request.graphtype == GT_TEXT)

int change_console(int i)
{
  return i;
}

struct tty_state
{
  unsigned short *textbuffer;	/* text buffer for virtual console	*/
  struct kgi_mode gr_mode;	/* graphics mode assigned		*/
  void *gr_buffer;		/* graphics buffer (user space !)	*/
};

void 
init_tty_state(struct tty_state *tty)
{
  memset(tty, 0, sizeof(struct tty_state));
  /* textbuffer, gr_mode, gr_buffer initialized */
}

static struct kgi_mode  textmode;
static struct tty_state tty[MAX_NR_CONSOLES];

#ifdef DEBUG

static void
printk_mode (struct kgi_mode *mode, char *msg)
{
  #define M(X) mode->X
  #define R(X) mode->request.X
  #define T(X) M(X.width), M(X.blankstart), M(X.syncstart), M(X.syncend),\
		M(X.blankend), M(X.total), M(X.polarity)
  printk(msg);
  printk("mode = {\n  {{ %i, %i},  /* visible */\n",
			R(visible.x),R(visible.y));
  printk("   { %i, %i},   /* virtual */\n",R(virtual.x),R(virtual.y));
  printk("   { %i, %i},   /* size [mm] */\n",R(size.x),R(size.y));
  printk("   %i,\n", R(graphtype));
  printk("   { %i, %i}},   /* textgrid */\n", R(textgrid.x),R(textgrid.y));
  printk("   { %i, %i, %i, %i, %i, %i, %i}, /* x timing */\n", T(x));
  printk("   { %i, %i, %i, %i, %i, %i, %i}, /* y timing */\n", T(y));
  printk("   { %i, %i},   /* magnify */\n", M(magnify.x), M(magnify.y));
  printk("   %x,          /* flags   */\n", M(flags));
  printk("   %i,          /* dclk    */\n", M(dclk));
  printk("   { %i, %i},   /* clk     */\n", M(clk.mul), M(clk.div));
  printk("   { %i, %i}    /* lclk    */\n}\n", M(lclk.mul), M(lclk.div));

  #undef M
  #undef R
  #undef T
}

#endif /* #ifdef DEBUG */

static int
kgi_check_mode(struct kgi_mode *mode, enum kgi_tm_cmd cmd)
{
  int cnt = 10;

  while (cnt-- && !(cmd == CMD_READY))
   {
     register int err;
     TRACE(printk("checkmode: cmd = %i; cnt = %i\n", cmd, cnt));
     if ((err = kgi_clock_check_mode  (mode, cmd)))  return err;
     if ((err = kgi_ramdac_check_mode (mode, cmd)))  return err;
     if ((err = kgi_chipset_check_mode(mode, cmd)))  return err;
     if ((err = kgi_monitor_check_mode(mode,&cmd)))  return err;
   }

  TRACE(printk_mode(mode, "mode checked and ok:\n"));
 
  return (++cnt) ? EOK : -EDRIVER;
}

static int
kgi_set_mode(struct kgi_mode *mode)
{
  int err;

  if ((err = kgi_check_mode (mode, CMD_CHECK)) )
    return err; 

  kgi_chipset_set_mode (mode);
  return EOK;
}

int
kgi_driver_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
		 unsigned long arg)
{
  register int err;
  int console = TTY_NR;			/* VC number of the process	*/

  switch(cmd)
   {
     case IOC_DRIVER_TESTDRIVER:	/* check if driver is present	*/
	return DEVICEFILE_MAGIC;

     case IOC_DRIVER_TESTMODE:		/* check if mode is possible	*/
	_out_buf = _io_buf;		/* _io_cnt is sizeof(ggi_mode)	*/
	return kgi_check_mode((struct kgi_mode *) _io_buf, CMD_PROPOSE);

     case IOC_DRIVER_SETGRAPHMODE:
	_out_buf = _io_buf;		/* _io_cnt is sizeof(ggi_mode)  */ 
	if ((err = kgi_check_mode((struct kgi_mode *) _io_buf, CMD_PROPOSE)))
	  return err;
	memcpy (&tty[console].gr_mode, _io_buf, sizeof(struct kgi_mode));

	/* !!! init if active console,
	 * ??? buffer reallocation ?
 	 */

	return EOK;

     case IOC_DRIVER_GETGRAPHMODE:	/* return graphics mode		*/
	if (NO_MODE_ASSIGNED(console))
	  return -EDRIVER;		/* no mode assigned */
	_out_buf = &(tty[console].gr_mode.request);
	return EOK;			/* _io_cnt is sizeof(ggi_mode)	*/

     case IOC_DRIVER_GRAPHICS:		/* switch to graphics console	*/
	if (NO_MODE_ASSIGNED(console))
	  return -EDRIVER;		/* no mode assigned */
	return change_console(console | VC_GRAPHMODE);

     case IOC_DRIVER_SETTEXTMODE:	/* set (console) text mode	*/
	if (((struct kgi_mode *) _io_buf)->request.graphtype != GT_TEXT)
	  {
	    TRACE(printk("ioctl(SETTEXTMODE): not a text mode.\n"));
	    return -EINVAL;
          }

	if ((err = kgi_check_mode((struct kgi_mode*) _io_buf, CMD_PROPOSE)))
	  return err;

	memcpy(&textmode, _io_buf, sizeof(textmode));

	/* !!!	buffer reallocation, 
	 * !!!	initialisation if active mode is textmode
	 * ???	other things ?
	 */
	_out_buf = &textmode;		/* _io_cnt is sizeof(ggi_mode)	*/
	return EOK;
	
     case IOC_DRIVER_GETTEXTMODE:	/* return (console) textmode	*/
	_out_buf = &textmode;		/* _io_cnt is sizeof(ggi_mode)	*/
	return EOK;

     case IOC_DRIVER_TEXT:		/* switch to text console	*/
	if (change_console(console & !VC_GRAPHMODE))
	  INTERNALERROR;		/* this should not happen. 	*/
	return EOK;			/* sould be always possible */

     default:
	return -EDRIVER;
   }
}

static void
kgi_done(void)
{
  kgi_clock_done();			/* deinitialisation, the calling     */
  kgi_ramdac_done();			/* sequence is reverse to initiali-  */
  kgi_monitor_done();			/* sation.			     */
  kgi_chipset_done();
}

static int
kgi_init(void)
{
  register int i;

  if (kgi_chipset_init())		/* chipset provides IO functions for */
    return -EIO;			/* all other parts, thus init first. */

  if (kgi_monitor_init())		/* chipset ok. Now init the monitor  */
    {					/* driver. If this fails (e.g. not   */
      kgi_chipset_done();		/* detected) deinitialize chipset.   */
      return -EIO;
    }

  if (kgi_ramdac_init())		/* Now go for the ramdac. 	     */
    {					
      kgi_monitor_done();		/* failed, thus deinitialize...	     */
      kgi_chipset_done();
      return -EIO;
    }

  if (kgi_clock_init())			/* last is the clock driver. 	     */
    {
      kgi_ramdac_done();		/* failed :-(			     */
      kgi_monitor_done();
      kgi_chipset_done();
      return -EIO;
    }

  for (i = 0; i < MAX_NR_CONSOLES; i++)	/* initialize tty states	     */
    init_tty_state(&(tty[i]));

  textmode.request = default_textmode;	/* now check the default textmode    */
  if (kgi_check_mode(&textmode, CMD_PROPOSE))
    {
      printk("fatal: kgi_check_mode(default_textmode) failed.\n");
      kgi_done();
      return -EIO;
    }

  kgi_set_mode(&textmode);
  return 0;				/* hardware should be ok now. :-))   */
}

/*\
|*|	interface stuff
\*/

typedef int (*ioctl_fn)(struct inode *inode, struct file *file,
			unsigned int cmd, unsigned long arg); 

static ioctl_fn ioctl[8] =
{
  kgi_driver_ioctl,	/* NOTE: this depends on the IOCTL definitions !!! */
  kgi_monitor_ioctl,
  kgi_ramdac_ioctl,
  kgi_clock_ioctl,
  kgi_chipset_ioctl,
  kgi_graphics_ioctl,
  NULL,
  NULL
};

unsigned int	_io_cnt;
unsigned char   _io_buf[IOCSIZE_MAX];
void *_out_buf;

static int
kgi_ioctl(struct inode *inode, struct file *file, unsigned int cmd, 
	  unsigned long arg)
{
  static int lock = 0;
  register int foo;
  int result = 0;

  if (lock) return -EBUSY;	/* ??? could this happen (new ioctl while  */ 
  lock = 1;			/* ??? the old one is not finished ?)	   */

  _out_buf = NULL;
  _io_cnt = (cmd & IOCSIZE_MASK) >> IOCSIZE_SHIFT;

  if (cmd & IOC_IN)		/* do we have input ?	*/
    {
      if ((foo = verify_area (VERIFY_READ, (void *) arg, _io_cnt)))
        { result = foo; goto unlock; }

      TRACE(printk("%s: copying %i bytes fromfs\n",__FILE__, _io_cnt));
      memcpy_fromfs (_io_buf, (void *) arg, _io_cnt);
    }

  foo = (cmd & IOC_SUBSYSTEM_MASK) >> IOC_SUBSYSTEM_SHIFT;
  if (ioctl[foo])
    result = ioctl[foo] (inode, file, cmd, arg);
   else
    {
      TRACE(printk("%s:%i: invalid subsystem (%i).\n",__FILE__, __LINE__, foo));
      result = -EINVAL;  goto unlock; 
    }

  if (cmd & IOC_OUT)
    {
      if ((foo = verify_area (VERIFY_WRITE, (void *) arg, _io_cnt)))
        { result = foo; goto unlock; }
      if (_out_buf && (0 < _io_cnt) && (_io_cnt < IOCSIZE_MAX))
        {
	  TRACE(printk("%s: copying %i bytes tofs\n",__FILE__, _io_cnt));
          memcpy_tofs ((void *) arg, _out_buf, _io_cnt);
        }
       else
        result = -EDRIVER;
    }

  unlock:
    lock = 0;
    TRACE(printk("%s: ioctl() = 0x%.8x\n", __FILE__, result));
    return result;
}

/* ??? Was machen kgi_graphmap_close und kgi_graphmap_nopage und 
   ??? wo werden die definiert (z.B. umbenennen in kgi_chipset_...
   ??? falls die vom chipset abhaengen.) */

void kgi_graphmap_close(void) {};
void kgi_graphmap_nopage(void) {};

struct vm_operations_struct
kgi_vmops =
{
	NULL,			/* open */
	kgi_graphmap_close,	/* close */
	NULL,			/* unmap */
	NULL,			/* protect */
	NULL,			/* sync */
	NULL,			/* advise */
	kgi_graphmap_nopage,	/* nopage */
	NULL,			/* wppage */
	NULL,			/* swapout */
	NULL,			/* swapin */
};

static int
kgi_mmap(struct inode *inode, struct file *file, struct vm_area_struct *vma)
{
/* ??? kann ich dass so einbauen ? */
#ifdef 0
   DEV_NUM_SETUP;

   if (vma->vm_offset & ~PAGE_MASK)
     return -ENXIO;

   #define VM(X) vma->vm_X

   switch(VM(offset))
    {
      case MMAP_LINEAR:
  	VM(inode) = inode;
 	VM(ops)   = &kgi_vmops;
	VM(offset)= (unsigned long) CurrState; /* This is ugly ... */
	inode->i_count++;

	/* Make sure paging tables are there ... */
 	zeromap_page_range (VM(start), VM(end) - VM(start), VM(page_prot));
	fast_unmap_page_range(VM(task), VM(start), VM(end) - VM(start));

	return 0;

      case MMAP_FRAME:
	VM(inode) = inode;
	inode->i_count++;

	/* Make sure paging tables are there ... */
	zeromap_page_range(VM(start), 0x10000, VM(page_prot));
	fast_remap_page_range(VM(task), VM(start), 0xa0000, 0x10000, VM(page_prot));

	return 0;

      default:		
	return -ENXIO;	/* We don't support qeer mappings right now! */
    }
#endif
}

static int
kgi_open(struct inode *inode, struct file *file)
{
#if 0
  int i;

/* ??? das sollte man irgendwie aendern. Wir koennten die 
   ??? minor IDs noch brauchen. */
  if (MINOR(inode->i_rdev) != 0xab)	/* MAGIC for ugly hack ... */
    return -ENODEV;
        
  if (!   (current->tty) 
       || (current->tty->device&0xff00) != 0x0400
/* ??? was bedeuted die 0x400 */
       || (current->tty->device&0xff)   >= 64)
    return -ENOTTY;			/* controlling tty is not a VT */

  i = 0;
  while( (i < MAX_CONS_NUM) && StateNow[i].available)
    i++;

  if (i >= MAX_CONS_NUM)
    return -EBUSY;

  MOD_INC_USE_COUNT;

  /* Go to FORGROUND-MODE and mark console as unavailable */
  StateNow[x].bgmode=0;
  StateNow[x].available = 0;
  StateNow[x].vt_num=current->tty->device&0xff;
  file->private_data=&StateNow[x];
#endif 
  return 0;				/* device ready to accept commands   */
}

static void
kgi_close(struct inode *inode, struct file *file)
{
#if 0
  DEV_NUM_SETUP;
  
  /* Return to the standard OpenState */
  
  if  (CurrState->bgmode==0) 
    _restorestate(&OpenState);		/* Need to set old state or */
  else 
    CurrState->bgmode=0;		/* it is already set ... */

  *CurrState=OpenState;
   CurrState->available = 1;

  MOD_DEC_USE_COUNT;
#endif
}

static struct file_operations 
kgi_fops =
{ NULL,		/* lseek	*/
  NULL,		/* read		*/
  NULL,         /* write	*/
  NULL,         /* readdir	*/
  NULL,		/* select	*/
  kgi_ioctl,	/* ioctl	*/
  kgi_mmap,	/* mmap		*/
  kgi_open,	/* open		*/
  kgi_close,	/* release	*/
  NULL		/* fsync	*/
};

/*\
|*|	module loading
\*/

int
init_module(void)
{
/*    unregister_chrdev(DEVICEFILE_MAJOR, DEVICEFILE_NAME); */

   if (register_chrdev(DEVICEFILE_MAJOR, DEVICEFILE_NAME, &kgi_fops))
     {
       printk("%s: can't register device file %s (major = %i).\n",\
	      __FILE__, DEVICEFILE_NAME, DEVICEFILE_MAJOR);
       return -EIO;
     } 

   if(kgi_init())
     {
       unregister_chrdev(DEVICEFILE_MAJOR, DEVICEFILE_NAME);
       return -EIO;
     }

   return 0;
}

void
cleanup_module (void)
{
  if (MOD_IN_USE)
    printk("kgi: scrdrv module is still in use. (not removed)\n");

  if (unregister_chrdev(DEVICEFILE_MAJOR, DEVICEFILE_NAME))
    printk("kgi: %s:%i: cleanup_module() failed.\n", __FILE__, __LINE__);
   else
    {
      kgi_done();
      printk("kgi module removed.\n");
    }
}

